//---------------------------------------------------------------------------
// File: TMainForm.cpp
//
//  TDx_Sound_Library
//  Example7 - Streaming Sounds or Music
//
// Date: 1st February, 2002
// Author: Darren John Dwyer
// Copyright: (c) 2002 Darren John Dwyer, All Rights Reserved.
//
// Description:
//
// This example "Streams" sound effects by creating a tiny buffer,
// then loading part of the music into the buffer whilst simultaneously
// playing another section of the buffer.
//
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "TMainForm.h"
//---------------------------------------------------------------------------
#pragma link "TDx_Sound"
#pragma link "TDx_SoundBuffer"
#pragma link "TDSCaps"
#pragma link "TDSBufferDesc"
#pragma link "TDx_Sound_Library_Install"
#pragma link "TDx_SoundNotify"
#pragma link "TDSBPositionNotify"
#pragma resource "*.dfm"
//---------------------------------------------------------------------------
TMainForm *MainForm;
//---------------------------------------------------------------------------
// This macro auto-indents messages, adds them to the status memo and
// truncates the memo as required to avoid exceptions creating huge numbers
// of message boxes saying "out of memory" or similar.
//---------------------------------------------------------------------------
int INDENT = 0;
//---------------------------------------------------------------------------
#define Msg(A)     try {\
                    AnsiString ssss = "";\
                    for (int nnnn=0;nnnn<INDENT*4;nnnn++) ssss = ssss + " ";\
                    ssss = ssss + (A);\
                    if (StatusMemo)\
                        if (StatusMemo->Lines)\
                            {\
                            if (StatusMemo->Lines->Text.Length()>60000) StatusMemo->Lines->Clear();\
                            StatusMemo->Lines->Add( ssss );\
                            }\
                    }\
                catch (Exception &E)\
                    {\
                    if (StatusMemo)\
                        if (StatusMemo->Lines)\
                            StatusMemo->Lines->Clear();\
                    }\
//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent* Owner)
    : TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormCreate(TObject *Sender)
{
Devices = new TStringList();
Modules = new TStringList();
Guids = new TList();

DevicesListBox->Items->Clear();
if (Dx_Sound1->DSEnumerate( NULL ))
    if (Devices->Count>0)
        {
        DevicesListBox->ItemIndex = 0;
        Dx_Sound1->Create( (GUID*) Guids->Items[DevicesListBox->ItemIndex] );
        }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormDestroy(TObject *Sender)
{
if (Dx_Sound1->Created) Dx_Sound1->Destroy();

if (Guids) delete Guids;
if (Modules) delete Modules;
if (Devices) delete Devices;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_Sound1Destroy(TObject *Sender)
{
if (Dx_SoundBuffer1->Created) Dx_SoundBuffer1->Destroy();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FileListBox1DblClick(TObject *Sender)
{
PlayButtonClick(this);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::PlayButtonClick(TObject *Sender)
{
Msg(" ");
Msg("<< Play Button was Clicked >>");
INDENT++;

if (Edit1->Text.Pos("*")!=0)
    {
    Msg( "    Cannot play *.WAV" );
    INDENT--;
    return;
    }

// if not already created, create a 2-second buffer to hold the streaming sound data
if (!Dx_SoundBuffer1->Created)
    {
    Msg("Creating From Stream( \""+Edit1->Text+"\" ) ...");

    if (Dx_SoundBuffer1->OpenStream( Edit1->Text ))
        {
        // use the format of the existing .wav file
        DSBufferDesc1->Format = Dx_SoundBuffer1->StreamFormat;

        // determine the # of bytes for a 2-seconds long buffer
        DSBufferDesc1->BufferBytes = Dx_SoundBuffer1->StreamFormat->nAvgBytesPerSec * 2;

        // create a 2s buffer of the existing format
        Dx_SoundBuffer1->Create( DSBufferDesc1, Dx_Sound1 );
        }
    }
else
    {
    // restart the buffer from the last streamed position.
    Msg("Restarting \""+LastFileName+"\" ...");

    if (Dx_SoundBuffer1->Playing)
        {
        Dx_SoundBuffer1->Stop();
        Dx_SoundBuffer1->SetCurrentPosition(0);
        }
    }

if (Dx_SoundBuffer1->Created)
    {
    // start streaming... note to set the buffer "LOOPING", so that we get repeated notifications
    Dx_SoundBuffer1->Play(0,DSBPLAY_LOOPING);
    if (Dx_SoundBuffer1->Playing) PlayingLabel->Caption = "Buffer is Streaming...";
    }
else
    {
    Msg("HERE, BUFFER NOT CREATED");
    }

INDENT--;
Msg("<< Play Button has Finished >>");
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_SoundBuffer1Create(TObject *Sender)
{
Msg("TDx_SoundBuffer was created ok");
INDENT++;

// setup buffer-related data
LastFileName = Edit1->Text;
FileNameLabel->Caption = "Playing: \""+Edit1->Text+"\"";
FileSizeLabel->Caption = "File Size: "+IntToStr(Dx_SoundBuffer1->StreamSize)+" bytes";
BufferSizeLabel->Caption = "Buffer Size: "+IntToStr(DSBufferDesc1->BufferBytes)+" bytes";
Msg(FileNameLabel->Caption);
Msg(FileSizeLabel->Caption);
Msg(BufferSizeLabel->Caption);
BytesRead = 0;
BytesToLoad = Dx_SoundBuffer1->StreamSize;

// update stats
Msg("BytesRead=="+IntToStr(BytesRead));
Msg("BytesToLoad=="+IntToStr(BytesToLoad));

// remember where we are
StopOnFirst = false;
StopOnSecond = false;
FinishedPlaying = false;

// create buffer notification positions
Msg("Creating Buffer Notifications");
Dx_SoundNotify1->Create( Dx_SoundBuffer1 );

INDENT--;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_SoundNotify1Create(TObject *Sender)
{
Msg("TDx_SoundNotify was created ok");
Msg("Setting Notification Positions");

DSBPositionNotify1->ArraySize = 3;
DSBPositionNotify1->Offset[0] = 0;
DSBPositionNotify1->Offset[1] = DSBufferDesc1->BufferBytes / 2;
DSBPositionNotify1->Offset[2] = DSBPN_OFFSETSTOP;

// create the notification positions
if (Dx_SoundNotify1->SetNotificationPositions( DSBPositionNotify1->ArraySize, DSBPositionNotify1 ))
    {
    Msg("Notification Positions set ok");

    PreloadFirstChunk();
    Msg("Pre-Loading First Chunk" );
    }
else
    Msg("Notification Positions not set!");
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::StopButtonClick(TObject *Sender)
{
Msg("");
Msg("<< Stop Button was Clicked >>");
INDENT++;

if (Dx_SoundBuffer1->Created)
    {
    Msg("About to destroy TDx_SoundBuffer");
    if (Dx_SoundBuffer1->Destroy()) PlayingLabel->Caption = "Buffer is Stopped.";
    }

INDENT--;
Msg("<< Stop Button has Finished >>");
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_SoundBuffer1Destroy(TObject *Sender)
{
Msg("TDx_SoundBuffer is being Destroyed..." );
INDENT++;

if (Dx_SoundBuffer1->Playing)
    {
    Msg("Stopping buffer playback..." );
    Dx_SoundBuffer1->Stop();
    }

if (Dx_SoundNotify1->Created)
    {
    Msg("About to Destroy TDx_SoundNotify");
    Dx_SoundNotify1->Destroy();
    }

if (Dx_SoundBuffer1->StreamIsOpen)
    {
    Msg("Closing the Stream");
    Dx_SoundBuffer1->CloseStream();
    }

INDENT--;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_SoundNotify1Destroy(TObject *Sender)
{
Msg("TDx_SoundNotify is being Destroyed..." );
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_SoundNotify1Notify(TObject *Sender, int Value)
{
Msg("");
Msg(">> Inside the TDx_SoundNotify::OnNotify() Event..." );
INDENT++;

// if the Dx_SoundNotify1::LocalNotify property is set to true,
// this event is called with the relevant notification position

// if the Dx_SoundNotify1::LocalNotify property is set to false,
// the relevant DSBPositionNotify[X]::OnNotify() events will be called

// either way, use the same actual code.... ;)

Msg("Handling Event #"+IntToStr(Value));
switch (Value)
    {
    case 0:    LoadSecondChunk(); break;
    case 1:    LoadFirstChunk(); break;
    case 2:    StopStreaming(); break;
    }

INDENT--;
Msg("");
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::QuitButtonClick(TObject *Sender)
{
Close();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DevicesListBoxClick(TObject *Sender)
{
if (Dx_Sound1->Created) Dx_Sound1->Destroy();
NotCreatedLabel->Visible = !Dx_Sound1->Create( (GUID*) Guids->Items[DevicesListBox->ItemIndex] );
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ToggleButtonClick(TObject *Sender)
{
Dx_SoundNotify1->LocalNotify = !Dx_SoundNotify1->LocalNotify;

if (Dx_SoundNotify1->LocalNotify) StatusLabel->Caption = "Events handled by TDx_SoundNotify::OnNotify()";
else StatusLabel->Caption = "Events handled by TDSBPositionNotify::OnNotify()";
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ClearButtonClick(TObject *Sender)
{
StatusMemo->Lines->Clear();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Dx_SoundBuffer1BufferLost(TObject *Sender)
{
Dx_SoundBuffer1->CreateFromFile( Edit1->Text, DSBufferDesc1, Dx_Sound1 );
Dx_SoundBuffer1->Play( 0,0 );
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_Sound1Create(TObject *Sender)
{
Dx_Sound1->SetCooperativeLevel( Application->Handle, DSSCL_PRIORITY );
}
//---------------------------------------------------------------------------



void __fastcall TMainForm::Dx_Sound1DSEnumerate(TObject *Sender,
      GUID *Guid, AnsiString DriverDescription, AnsiString DriverModule,
      void *Context, bool &Finished)
{
if (Devices && Modules && Guids)
    {
    Guids->Add( (void*) Guid );
    Modules->Add( DriverModule );
    Devices->Add( DriverDescription );

    DevicesListBox->Items->Add( DriverDescription + ", ("+DriverModule+")" );
    }
else
    Finished = true;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Dx_Sound1Error(TObject *Sender,
      AnsiString Function, AnsiString Error, AnsiString ErrorMessage)
{
Msg("TDx_Draw Error: "+Error+", "+Function+", "+ErrorMessage);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Dx_SoundBuffer1Error(TObject *Sender,
      AnsiString Function, AnsiString Error, AnsiString ErrorMessage)
{
Msg("Dx_SoundBuffer1 Error: "+Error+", "+Function+", "+ErrorMessage);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DSBufferDesc1Error(TObject *Sender,
      AnsiString Function, AnsiString Error, AnsiString ErrorMessage)
{
Msg("DSBufferDesc1 Error: "+Error+", "+Function+", "+ErrorMessage);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Dx_SoundNotify1Error(TObject *Sender,
      AnsiString Function, AnsiString Error, AnsiString ErrorMessage)
{
Msg("Dx_SoundNotify1 Error: "+Error+", "+Function+", "+ErrorMessage);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DSBPositionNotify1Error(TObject *Sender,
      AnsiString Function, AnsiString Error, AnsiString ErrorMessage)
{
Msg("DSBPositionNotify1 Error: "+Error+", "+Function+", "+ErrorMessage);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DSBPositionNotify1Notify(TObject *Sender,
      int Notification)
{
Msg("In TDSBPositionNotify::OnNotify("+IntToStr(Notification)+")" );
Dx_SoundNotify1Notify( Sender, Notification );
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::PreloadFirstChunk()
{
// read the initial chunk of sound data into the first half of the buffer
byte* first_chunk_addr;
dword first_chunk_size;
byte* second_chunk_addr;
dword second_chunk_size;
dword bytes_actually_read = 0;

// lock the first half of the buffer
dword bytes_to_lock = DSBufferDesc1->BufferBytes/2;

if (Dx_SoundBuffer1->Lock( 0, bytes_to_lock,
                            (void**) &first_chunk_addr, &first_chunk_size,
                            (void**) &second_chunk_addr, &second_chunk_size, 0 ))
    {
    Msg("First half of Buffer was locked");

    // read data from the stream into the first half of the buffer
    if (Dx_SoundBuffer1->ReadFromStream( first_chunk_size, first_chunk_addr, &bytes_actually_read ))
        {
        Msg("First half of Buffer was read from file, "+IntToStr(bytes_actually_read)+" bytes");

        BytesRead += bytes_actually_read;
        BytesToLoad -= bytes_actually_read;

        // check whether all data was read correctly.
        if (bytes_actually_read < first_chunk_size)
            {
            FinishedPlaying = true;
            StopOnSecond = true;
            Msg("End of Stream found, Stopping on Second Chunk" );
            Msg("bytes_actually_read["+IntToStr(bytes_actually_read)+"] < first_chunk_size["+IntToStr(first_chunk_size)+"]");

            // leave just-loaded sound data and fill remainder of the chunk with zero's
            Msg("Filling from offset: "+IntToStr(((dword) &first_chunk_addr[bytes_actually_read]) - (dword) first_chunk_addr)+" for "+
                IntToStr( first_chunk_size-bytes_actually_read ) + " bytes" );
            FillMemory( &first_chunk_addr[bytes_actually_read], first_chunk_size-bytes_actually_read, 0 );
            }

        }
    else
        {
        // an error occurred during the ReadFromStream()
        StopOnFirst = true;
        FinishedPlaying = true;
        }

    // finished with the buffer
    Dx_SoundBuffer1->Unlock( (void*) first_chunk_addr, first_chunk_size,
                             (void*) second_chunk_addr, second_chunk_size );
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::LoadSecondChunk()
{
Msg("");
Msg(">> Inside the LoadSecondChunk() Method..." );
if (FinishedPlaying && StopOnFirst)
    {
    Msg("Stopping Stream");
    Dx_SoundBuffer1->Stop();
    StopButtonClick(this);
    return;
    }
INDENT++;

NotificationLabel->Caption = "Notification: Start";
PlayChunkLabel->Caption = "Playing: Chunk #0";
LoadChunkLabel->Caption = "Loading: Chunk #1";

Msg(NotificationLabel->Caption);
Msg(PlayChunkLabel->Caption);
Msg(LoadChunkLabel->Caption);

byte* first_chunk_addr;
dword first_chunk_size;
byte* second_chunk_addr;
dword second_chunk_size;
dword bytes_actually_read = 0;

dword bytes_to_lock = DSBufferDesc1->BufferBytes-DSBPositionNotify1->Offset[1];

if (Dx_SoundBuffer1->Lock( DSBPositionNotify1->Offset[1], bytes_to_lock,
                            (void**) &first_chunk_addr, &first_chunk_size,
                            (void**) &second_chunk_addr, &second_chunk_size, 0 ))
    {
    Msg("Second half of Buffer was locked, "+IntToStr(bytes_to_lock)+" bytes");

    if (FinishedPlaying && StopOnSecond)
        {
        FillMemory( first_chunk_addr, first_chunk_size, 0 );
        }
    else
        {
        dword bytes_to_read = first_chunk_size;
        if (bytes_to_read > BytesToLoad) bytes_to_read = BytesToLoad;
        if (Dx_SoundBuffer1->ReadFromStream( bytes_to_read, first_chunk_addr, &bytes_actually_read ))
            {
            Msg("Second half of Buffer was read from file, "+IntToStr(bytes_actually_read)+" bytes");
            BytesRead += bytes_actually_read;
            BytesToLoad -= bytes_actually_read;

            if (bytes_actually_read < first_chunk_size)
                {
                FinishedPlaying = true;
                StopOnFirst = true;

                Msg("End of Stream found, Stopping on First" );
                Msg("bytes_actually_read["+IntToStr(bytes_actually_read)+"] < first_chunk_size["+IntToStr(first_chunk_size)+"]");
                Msg("Filling from offset: "+IntToStr(((dword) &first_chunk_addr[bytes_actually_read]) - (dword) first_chunk_addr)+" for "+
                    IntToStr( first_chunk_size-bytes_actually_read ) + " bytes" );
                FillMemory( &first_chunk_addr[bytes_actually_read], first_chunk_size-bytes_actually_read, 0 );
                }

            }
        else
            {
            StopOnFirst = true;
            FinishedPlaying = true;
            }
        }
    Dx_SoundBuffer1->Unlock( (void*) first_chunk_addr, first_chunk_size,
                             (void*) second_chunk_addr, second_chunk_size );
    }

BytesReadLabel->Caption = "BytesRead: "+IntToStr(BytesRead)+" bytes";
BytesToLoadLabel->Caption = "BytesToLoad: "+IntToStr(BytesToLoad)+" bytes";
Msg(BytesReadLabel->Caption);
Msg(BytesToLoadLabel->Caption);
INDENT--;
Msg("");
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::LoadFirstChunk()
{
Msg("");
Msg(">> Inside the LoadFirstChunk() method ..." );
if (FinishedPlaying && StopOnSecond)
    {
    Msg("Stopping Stream");
    Dx_SoundBuffer1->Stop();
    StopButtonClick(this);
    return;
    }
INDENT++;

NotificationLabel->Caption = "Notification: Middle";
PlayChunkLabel->Caption = "Playing: Chunk #1";
LoadChunkLabel->Caption = "Loading: Chunk #0";

Msg(NotificationLabel->Caption);
Msg(PlayChunkLabel->Caption);
Msg(LoadChunkLabel->Caption);

byte* first_chunk_addr;
dword first_chunk_size;
byte* second_chunk_addr;
dword second_chunk_size;
dword bytes_actually_read = 0;

dword bytes_to_lock = DSBPositionNotify1->Offset[1]-DSBPositionNotify1->Offset[0];

if (Dx_SoundBuffer1->Lock( DSBPositionNotify1->Offset[0], bytes_to_lock,
                            (void**) &first_chunk_addr, &first_chunk_size,
                            (void**) &second_chunk_addr, &second_chunk_size,0 ))
    {
    Msg("First half of Buffer was locked, "+IntToStr(bytes_to_lock)+" bytes");
    if (FinishedPlaying && StopOnFirst)
        {
        FillMemory( first_chunk_addr, first_chunk_size, 0 );
        }
    else
        {
        dword bytes_to_read = first_chunk_size;
        if (bytes_to_read > BytesToLoad) bytes_to_read = BytesToLoad;
        if (Dx_SoundBuffer1->ReadFromStream( bytes_to_read, first_chunk_addr, &bytes_actually_read ))
            {
            Msg("First half of Buffer was read from file, "+IntToStr(bytes_actually_read)+" bytes");
            BytesRead += bytes_actually_read;
            BytesToLoad -= bytes_actually_read;

            if (bytes_actually_read < first_chunk_size)
                {
                FinishedPlaying = true;
                StopOnSecond = true;

                Msg("End of Stream found, Stopping On Second" );
                Msg("bytes_actually_read["+IntToStr(bytes_actually_read)+"] < first_chunk_size["+IntToStr(first_chunk_size)+"]");
                Msg("Filling from offset: "+IntToStr(((dword) &first_chunk_addr[bytes_actually_read]) - (dword) first_chunk_addr)+" for "+
                    IntToStr( first_chunk_size-bytes_actually_read ) + " bytes" );
                FillMemory( &first_chunk_addr[bytes_actually_read], first_chunk_size-bytes_actually_read, 0 );
                }
            }
        else
            {
            StopOnSecond = true;
            FinishedPlaying = true;
            }
        }

    Dx_SoundBuffer1->Unlock( (void*) first_chunk_addr, first_chunk_size,
                             (void*) second_chunk_addr, second_chunk_size );
    }

BytesReadLabel->Caption = "BytesRead: "+IntToStr(BytesRead)+" bytes";
BytesToLoadLabel->Caption = "BytesToLoad: "+IntToStr(BytesToLoad)+" bytes";
Msg(BytesReadLabel->Caption);
Msg(BytesToLoadLabel->Caption);
INDENT--;
Msg("");
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::StopStreaming()
{
Msg("");
Msg(">> Inside the StopStreaming() method..." );
INDENT++;

NotificationLabel->Caption = "Notification: OFFSETSTOP";
Msg(NotificationLabel->Caption);

Msg("Stopping Stream");
StopButtonClick(this);

INDENT--;
Msg("");
}
//---------------------------------------------------------------------------

